;This TSR redirects INT40 or Int13 for transferring floppy data not into UMB
;For use in conjunction with UMBPCI
;This little program itself must be loaded into low memory! (no LH)
;Possible Enhancements:
;- make a dual-mode, single driver (.EXE, for simpler usage)
;- keep sector only in conventional memory; hook code can reside in UMB
;- make driver uninstallable (not applicable for a .SYS driver section)

; To BUILD this version, use either:
; nasm -DMAKECOM -o lowdma.com lowdma.nas
; or for the device driver version:
; nasm -o lowdma.sys lowdma.nas
; You need no other tools than the free NASM assembler :-)


CODESEG

%IFDEF MAKECOM
	; original .com bufstart was 5ch
bufstart	EQU	60h
	org	100h
.start:
comstart:
	;This transient code will be overwritten by next sector...
	call	transient
	jc	@@comstart_e
	mov	es,[2Ch]	;environment
	mov	ah,49h
	int	21h		;free memory block
	mov	dx,ISRE
	int	27h		;remain resident
@@comstart_e:	retn
%ELSE

%DEFINE	comstart	()

struc	INITREQUEST
 .irLength:	resb	1
 .irUnit:	resb	1
 .irFunction:	resb	1	;function 00 = init is used
 .irStatus:	resw	1
 .irReserved:	resb	8 
 .irUnits:	resb	1
 .irEndAddress:	resd	1	;highest resident memory location
 .irParamAddress:	resd	1	;command line (unused)
 .irDriveNumber:	resb	1
 .irMessageFlag:	resb	1
endstruc

DNext	dd -1		;address of next driver or -1 if end of chain
DAttr	dw 0E000h	;attribute
DStrat	dw strat	;offset of "strategy" routine
DIntr	dw intr	;offset of "interrupt" routine
DName	db 'LOWDMA$$'	;internal name of .SYS
	; original .sys bufstart was 12h
bufstart	EQU	20h		;after the name
ParamAddr	dd	0
strat:
	mov	word [ParamAddr],bx
	mov	word [ParamAddr+2],es
	retf
intr:
	push	es
	push	bx
	push	ax
	pushf
	 les	bx,[ParamAddr]
	 mov	WORD [es:bx + INITREQUEST.irStatus],8103h	;an error code
	 mov	word [es:bx + INITREQUEST.irEndAddress],0
	 mov	word [es:bx + INITREQUEST.irEndAddress+2],cs
	 cmp	BYTE [es:bx + INITREQUEST.irFunction],0
	 jnz	@@intr_e
	 push	es
	 push	bx
	  call	transient
	 pop	bx
	 pop	es
	 jc	@@intr_e
	 mov	word [es:bx + INITREQUEST.irEndAddress],ISRE
	 mov	WORD [es:bx + INITREQUEST.irStatus],100h	;the OK code
@@intr_e:	popf
	pop	ax
	pop	bx
	pop	es
	retf

%ENDIF

transient:
	mov	dx,msg$
	mov	ah,9
	int	21h		;write something
	push	sp
	pop	ax		;check for (186 or) 286
	cmp	ax,sp
	mov	dx,no286$
	jc	@@transient_err
	CPU	286			;needed for PUSHA/POPA and SHR n
	mov	ax,cs
	cmp	ah,0A0h
	mov	dx,err$
	cmc			; ah must be below a0 to be okay
	jc	@@transient_err
	push	ax
	add	ax, (bufstart+15) / 16
	and	ax,0fffh	; check position of buffer relative to DMA boundaries
	cmp	ax,0fc0h	; in the last 1k of a DMA block?
	pop	ax
	mov	dx,err2$
	cmc			; ax (offset) must be below fc0(0) to be okay
	jc	@@transient_err
	; Now we are ready to install, try Int40 first
;	mov	ax,3540h
;	int	21h
;	mov	dx,es
;	or	dx,bx		;Got a null pointer?
;	jnz	@@transient_set
	mov	ax,3513h
	int	21h		;get interrupt vector
@@transient_set:
	mov	word [OldInt13],bx
	mov	word [OldInt13+2],es
	mov	dx,NewInt13
	mov	ah,25h
	int	21h		;set interrupt vector
	mov	dx,ok$
@@transient_err:
	pushf
	 mov	ah,9
	 int	21h
	popf
	retn

msg$	db	'LowDMA (haftmann#software 11/01, DMA fix 12/06): $'
no286$	db	'requires 80286 or higher processor!',13,10,'$'
err$	db	'must not reside in UMB!',13,10,'$'
err2$	db	'must not be loaded at a 64k DMA boundary!',13,10,'$'
ok$	db	'installed :-)'
%IFDEF MAKECOM
	db	' (no uninstall)'
%ENDIF
	db	13,10,'$'

%IFDEF MAKECOM
	times	(bufstart+512-$+$$)-100h	db 0
%ELSE
	times	bufstart+512-$+$$		db 0
%ENDIF
ISRA:
NewInt13:
	pushf
	 cmp	ah,2		;read sectors?
	 je	@@NewInt13_r
	 cmp	ah,3		;write sectors?
	 jne	@@NewInt13_no
@@NewInt13_r:	 or	al,al		;let BIOS handle this error condition
	 jz	@@NewInt13_no
	 test	dl,80h		;for hard disk?
	 jnz	@@NewInt13_no
	 push	ax
	 push	bx
	push cx
	mov	cl,al
	mov	ch,0
	shl	cx,(9-4)	; CX! convert unit from 0.5k sectors to paragraphs
	  mov	ax,es
	  shr	bx,4		;discard lower 4 bits
	  add	ax,bx		; AX = bufferstart/16: HMA? Wraps to conventional memory!
	inc	cx
	mov	bx,cx		; transfer size in paras
	pop cx
	
	  cmp	ah,0A0h		;UMB or ROM? No->CY=1
	jnc @@NewInt13_h	; UMB is always bad
	and	ax,0fffh	; look at DMA boundary alignment
	add	bx,ax		; find end of transfer block, modulo 64k
	test	bh,0f0h		; wraps into next block?
	clc			; assume "bad", so clear the carry flag
	jnz @@NewInt13_h	; indeed, we found a bad wrap
	stc			; set carry to indicate "not bad"
@@NewInt13_h:
	pop	bx
	pop	ax
	 jnc	@@NewInt13_rw		;redirect if UMB or wrap
@@NewInt13_no:
	popf
	db	0EAh		;jump far
OldInt13 dd	0

@@NewInt13_rw:	popf		; the wrapper handler to avoid UMBs
	pusha
	push	ds
	push	es
	 mov	bp,sp
	 push	bx
	 push	cx
	 push	ax
	 push	cs
	 push	bufstart	;one byte instead of "offset buffer"
	 shr	byte [bp+18h],1	;shift carry out
	 mov	byte [bp+12h],0	;number of processed sectors ->AL
	 cld
@@NewInt13_l:
	 lds	si,[bp-2]	;source data
	 les	di,[bp-10]	;destination buffer
	 mov	cx,256
	 rep	movsw		;copy data

	 les	bx,[bp-10]
	 mov	cx,[bp-4]
	 mov	ah,[bp-5]
	 mov	al,1
	 sti
	 pushf
	 call	FAR [cs:OldInt13]	;issue Int13 with local buffer reference

	 lds	si,[bp-10]	;source buffer
	 les	di,[bp-2]	;destination data
	 mov	cx,256
	 rep	movsw		;copy data back

	 jc	@@NewInt13_e		;on error: leave!
	 add	byte [bp-1],2	;same as adding 512 to a word
	 inc	byte [bp+12h]	;sectors transferred ->AL
	 inc	byte [bp-4]	;next sector (no calculation of head jump)
	 dec	byte [bp-6]	;residual sector count
	 jnz	@@NewInt13_l
	 clc
@@NewInt13_e:
	 mov	[bp+13h],ah	;save status ->AH
	 rcl	byte [bp+18h],1	;shift carry in
	mov	sp,bp
	pop	es
	pop	ds
	popa
	iret
ISRE:
	end
